@@ -18,6 +18,7 @@
 #include <errno.h>
 #include <pwd.h>
 #include <grp.h>
+#include <stddef.h>
 #ifdef HAVE_SYS_XATTR_H
 #include <sys/xattr.h>
 #endif
@@ -38,9 +39,12 @@
 #undef PACKAGE_TARNAME
 #undef PACKAGE_VERSION
 #include <gfarm/gfarm.h>
+#include <gfarm2fs.h>
+#include <replicate.h>
+#include "gfarm2fs_msg_enums.h"
 
-/* for better scalability */
-#define USE_GETDIR
+/* for old interface */
+#undef USE_GETDIR
 
 /* XXX FIXME */
 #define GFS_DEV		((dev_t)-1)
@@ -49,19 +53,107 @@
 
 char *program_name = "gfarm2fs";
 
+static char GFARM2FS_SYSLOG_FACILITY_DEFAULT[] = "local0";
+static char GFARM2FS_SYSLOG_PRIORITY_DEFAULT[] = "notice";
+static char GFARM2FS_SYSLOG_PRIORITY_DEBUG[] = "debug";
+
+static const char *mount_point;
+
+#define PATH_LEN_LIMIT 200
+static char *syslog_fmt = "<%s:%s>[%s]%s%s: %s";
+static char *trunc_str = "(...)";
+
+#define gfarm2fs_check_error(msgNo, fuse_opname, gfarm_funcname, \
+			     gfarm_path, gfarm_e) \
+{ \
+	if (gfarm_e != GFARM_ERR_NO_ERROR) { \
+		int ret_errno    = gfarm_error_to_errno(gfarm_e); \
+		int path_len     = strlen(gfarm_path); \
+		int path_offset  = 0; \
+		char *path_prefix = ""; \
+		if (path_len > PATH_LEN_LIMIT) { \
+			path_offset = path_len - PATH_LEN_LIMIT; \
+			path_prefix = trunc_str; \
+		} \
+		if (ret_errno == EINVAL || fuse_opname == OP_RELEASE) { \
+			gflog_error(msgNo, syslog_fmt, fuse_opname, \
+				gfarm_funcname, mount_point, \
+				path_prefix, gfarm_path + path_offset, \
+				gfarm_error_string(gfarm_e)); \
+		} else if (ret_errno != ENOENT) { \
+			gflog_info(msgNo, syslog_fmt, fuse_opname, \
+				gfarm_funcname, mount_point, \
+				path_prefix, gfarm_path + path_offset, \
+				gfarm_error_string(gfarm_e)); \
+		} else { \
+			gflog_debug(msgNo, syslog_fmt, fuse_opname, \
+				gfarm_funcname, mount_point, \
+				path_prefix, gfarm_path + path_offset, \
+				gfarm_error_string(gfarm_e)); \
+		} \
+	} \
+}
+
+static char OP_GETATTR[] = "GETATTR";
+static char OP_FGETATTR[] = "FGETATTR";
+#if 0 /* XXX Part of invoking gfs_access() is defined "if 0" now */
+static char OP_ACCESS[] = "ACCESS";
+#endif
+static char OP_READLINK[] = "READLINK";
+#ifndef USE_GETDIR
+static char OP_OPENDIR[] = "OPENDIR";
+static char OP_READDIR[] = "READDIR";
+static char OP_RELEASEDIR[] = "RELEASEDIR";
+#else /* USE_GETDIR */
+static char OP_GETDIR[] = "GETDIR";
+#endif /* USE_GETDIR */
+static char OP_MKNOD[] = "MKNOD";
+static char OP_MKDIR[] = "MKDIR";
+static char OP_UNLINK[] = "UNLINK";
+static char OP_RMDIR[] = "RMDIR";
+static char OP_SYMLINK[] = "SYMLINK";
+static char OP_RENAME[] = "RENAME";
+static char OP_LINK[] = "LINK";
+static char OP_CHMOD[] = "CHMOD";
+static char OP_CHOWN[] = "CHOWN";
+static char OP_TRUNCATE[] = "TRUNCATE";
+static char OP_FTRUNCATE[] = "FTRUNCATE";
+static char OP_UTIME[] = "UTIME";
+static char OP_CREATE[] = "CREATE";
+static char OP_OPEN[] = "OPEN";
+static char OP_READ[] = "READ";
+static char OP_WRITE[] = "WRITE";
+static char OP_STATFS[] = "STATFS";
+static char OP_RELEASE[] = "RELEASE";
+static char OP_FSYNC[] = "FSYNC";
+#if defined(HAVE_SYS_XATTR_H) && defined(ENABLE_XATTR)
+static char OP_SETXATTR[] = "SETXATTR";
+static char OP_GETXATTR[] = "GETXATTR";
+static char OP_LISTXATTR[] = "LISTXATTR";
+static char OP_REMOVEXATTR[] = "REMOVEXATTR";
+#endif /* HAVE_SYS_XATTR_H && ENABLE_XATTR */
+
 static uid_t
 get_uid(char *user)
 {
 	struct passwd *pwd;
+	char *luser;
 
 	if (strcmp(gfarm_get_global_username(), user) == 0)
-		return getuid(); /* my own file */
+		return (getuid()); /* my own file */
 
-	/* assumes that the same username exists on the local system */
-	if ((pwd = getpwnam(user)) != NULL)
-		return pwd->pw_uid;
-
-	/* XXX FIXME - some other's file */
+	/*
+	 * XXX - this interface will be changed soon to support
+	 * multiple gfmds.
+	 */
+	if (gfarm_global_to_local_username(user, &luser)
+	    == GFARM_ERR_NO_ERROR) {
+		pwd = getpwnam(luser);
+		free(luser);
+		if (pwd != NULL)
+			return (pwd->pw_uid);
+	}
+	/* cannot conver to a local account */
 	return (0);
 }
 
@@ -69,12 +161,20 @@ static int
 get_gid(char *group)
 {
 	struct group *grp;
+	char *lgroup;
 
-	/* assumes that the same groupname exists on the local system */
-	if ((grp = getgrnam(group)) != NULL)
-		return grp->gr_gid;
-
-	/* XXX FIXME */
+	/*
+	 * XXX - this interface will be changed soon to support
+	 * multiple gfmds.
+	 */
+	if (gfarm_global_to_local_groupname(group, &lgroup)
+	    == GFARM_ERR_NO_ERROR) {
+		grp = getgrnam(lgroup);
+		free(lgroup);
+		if (grp != NULL)
+			return (grp->gr_gid);
+	}
+	/* cannot conver to a local group */
 	return (0);
 }
 
@@ -110,8 +210,11 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 	gfarm_error_t e;
 
 	e = gfs_lstat_cached(path, &st);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000001, OP_GETATTR,
+					"gfs_lstat_cached", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	copy_gfs_stat(stbuf, &st);
 	gfs_stat_free(&st);
@@ -132,8 +235,11 @@ gfarm2fs_fgetattr(const char *path, struct stat *stbuf,
 	gfarm_error_t e;
 
 	e = gfs_pio_stat(get_filep(fi), &st);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000002, OP_FGETATTR,
+					"gfs_pio_stat", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	copy_gfs_stat(stbuf, &st);
 	gfs_stat_free(&st);
@@ -149,6 +255,8 @@ gfarm2fs_access(const char *path, int mask)
 	gfarm_error_t e;
 
 	e = gfs_access(path, mask);
+	gfarm2fs_check_error(GFARM_MSG_2000003, OP_ACCESS,
+				"gfs_access", path, e);
 	return (-gfarm_error_to_errno(e));
 #endif
 }
@@ -161,8 +269,11 @@ gfarm2fs_readlink(const char *path, char *buf, size_t size)
 	size_t len;
 
 	e = gfs_readlink(path, &src);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000004, OP_READLINK,
+					"gfs_readlink", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	len = strlen(src);
 	if (len >= size)
@@ -180,8 +291,11 @@ gfarm2fs_opendir(const char *path, struct fuse_file_info *fi)
 	GFS_Dir dp;
 
 	e = gfs_opendir_caching(path, &dp);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000005, OP_OPENDIR,
+					"gfs_opendir_caching", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	fi->fh = (unsigned long) dp;
 	return (0);
@@ -214,6 +328,8 @@ gfarm2fs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
 		if (filler(buf, de->d_name, &st, 0))
 			break;
 	}
+	gfarm2fs_check_error(GFARM_MSG_2000006, OP_READDIR,
+				"gfs_readdir", path, e);
 
 	return (0);
 }
@@ -222,9 +338,12 @@ static int
 gfarm2fs_releasedir(const char *path, struct fuse_file_info *fi)
 {
 	GFS_Dir dp = get_dirp(fi);
+	gfarm_error_t e;
 
 	(void) path;
-	gfs_closedir(dp);
+	e = gfs_closedir(dp);
+	gfarm2fs_check_error(GFARM_MSG_2000007, OP_RELEASEDIR,
+				"gfs_closedir", path, e);
 	return (0);
 }
 #else /* USE_GETDIR */
@@ -237,15 +356,24 @@ gfarm2fs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
 	struct gfs_dirent *de;
 
 	e = gfs_opendir_caching(path, &dp);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000008, OP_GETDIR,
+					"gfs_opendir_caching", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	while ((e = gfs_readdir(dp, &de)) == GFARM_ERR_NO_ERROR &&
 		de != NULL) {
 		if (filler(h, de->d_name, de->d_type << 12, de->d_fileno))
 			break;
 	}
+	gfarm2fs_check_error(GFARM_MSG_2000009, OP_GETDIR,
+				"gfs_readdir", path, e);
+
 	e2 = gfs_closedir(dp);
+	gfarm2fs_check_error(GFARM_MSG_2000010, OP_GETDIR,
+				"gfs_closedir", path, e2);
+
 	if (e == GFARM_ERR_NO_ERROR)
 		e = e2;
 
@@ -264,8 +392,15 @@ gfarm2fs_mknod(const char *path, mode_t mode, dev_t rdev)
 
 	e = gfs_pio_create(path, GFARM_FILE_WRONLY, mode & GFARM_S_ALLPERM,
 		&gf);
-	if (e == GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000011, OP_MKNOD,
+					"gfs_pio_create", path, e);
+	} else {
 		e = gfs_pio_close(gf);
+		gfarm2fs_check_error(GFARM_MSG_2000012, OP_MKNOD,
+					"gfs_pio_close", path, e);
+	}
+
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -275,6 +410,8 @@ gfarm2fs_mkdir(const char *path, mode_t mode)
 	gfarm_error_t e;
 
 	e = gfs_mkdir(path, mode & GFARM_S_ALLPERM);
+	gfarm2fs_check_error(GFARM_MSG_2000013, OP_MKDIR,
+				"gfs_mkdir", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -284,6 +421,8 @@ gfarm2fs_unlink(const char *path)
 	gfarm_error_t e;
 
 	e = gfs_unlink(path);
+	gfarm2fs_check_error(GFARM_MSG_2000014, OP_UNLINK,
+				"gfs_unlink", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -293,6 +432,8 @@ gfarm2fs_rmdir(const char *path)
 	gfarm_error_t e;
 
 	e = gfs_rmdir(path);
+	gfarm2fs_check_error(GFARM_MSG_2000015, OP_RMDIR,
+				"gfs_rmdir", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -302,6 +443,8 @@ gfarm2fs_symlink(const char *from, const char *to)
 	gfarm_error_t e;
 
 	e = gfs_symlink(from, to);
+	gfarm2fs_check_error(GFARM_MSG_2000016, OP_SYMLINK,
+				"gfs_symlink", to, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -311,6 +454,8 @@ gfarm2fs_rename(const char *from, const char *to)
 	gfarm_error_t e;
 
 	e = gfs_rename(from, to);
+	gfarm2fs_check_error(GFARM_MSG_2000017, OP_RENAME,
+				"gfs_rename", from, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -320,6 +465,8 @@ gfarm2fs_link(const char *from, const char *to)
 	gfarm_error_t e;
 
 	e = gfs_link(from, to);
+	gfarm2fs_check_error(GFARM_MSG_2000018, OP_LINK,
+				"gfs_link", to, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -329,34 +476,44 @@ gfarm2fs_chmod(const char *path, mode_t mode)
 	gfarm_error_t e;
 
 	e = gfs_chmod(path, mode & GFARM_S_ALLPERM);
+	gfarm2fs_check_error(GFARM_MSG_2000019, OP_CHMOD,
+				"gfs_chmod", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
+/* returned string should be free'ed if it is not NULL */
 static char *
 get_user(uid_t uid)
 {
 	struct passwd *pwd;
+	char *guser;
 
 	if (uid == getuid())
-		return gfarm_get_global_username();
+		return (strdup(gfarm_get_global_username()));
 
-	/* assumes that the same username exists on the gfarm filesystem */
-	if ((pwd = getpwuid(uid)) != NULL)
-		return pwd->pw_name;
+	/* use the user map file to identify the global user */
+	if ((pwd = getpwuid(uid)) != NULL &&
+	    gfarm_local_to_global_username(pwd->pw_name, &guser)
+	    == GFARM_ERR_NO_ERROR)
+		return (guser);
 
-	return NULL;
+	return (NULL);
 }
 
+/* returned string should be free'ed if it is not NULL */
 static char *
-get_group(uid_t gid)
+get_group(gid_t gid)
 {
 	struct group *grp;
+	char *ggroup;
 
-	/* assumes that the same groupname exists on the gfarm filesystem */
-	if ((grp = getgrgid(gid)) != NULL)
-		return grp->gr_name;
+	/* use the group map file to identify the global group */
+	if ((grp = getgrgid(gid)) != NULL &&
+	    gfarm_local_to_global_groupname(grp->gr_name, &ggroup)
+	    == GFARM_ERR_NO_ERROR)
+		return (ggroup);
 
-	return NULL;
+	return (NULL);
 }
 
 static int
@@ -368,14 +525,22 @@ gfarm2fs_chown(const char *path, uid_t uid, gid_t gid)
 	if (uid == -1)
 		user = NULL;
 	else if ((user = get_user(uid)) == NULL)
-		return EINVAL;
+		return (-EPERM);
 
 	if (gid == -1)
 		group = NULL;
-	else if ((group = get_group(gid)) == NULL)
-		return EINVAL;
-
+	else if ((group = get_group(gid)) == NULL) {
+		if (user != NULL)
+			free(user);
+		return (-EPERM);
+	}
 	e = gfs_chown(path, user, group);
+	gfarm2fs_check_error(GFARM_MSG_2000020, OP_CHOWN,
+				"gfs_chown", path, e);
+	if (user != NULL)
+		free(user);
+	if (group != NULL)
+		free(group);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -386,23 +551,33 @@ gfarm2fs_truncate(const char *path, off_t size)
 	GFS_File gf;
 
 	e = gfs_pio_open(path, GFARM_FILE_WRONLY, &gf);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000021, OP_TRUNCATE,
+					"gfs_pio_open", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	e = gfs_pio_truncate(gf, size);
+	gfarm2fs_check_error(GFARM_MSG_2000022, OP_TRUNCATE,
+				"gfs_pio_truncate", path, e);
 	e2 = gfs_pio_close(gf);
+	gfarm2fs_check_error(GFARM_MSG_2000023, OP_TRUNCATE,
+				"gfs_pio_close", path, e2);
+	gfarm2fs_replicate(path, NULL);
 
 	return (-gfarm_error_to_errno(e != GFARM_ERR_NO_ERROR ? e : e2));
 }
 
 static int
 gfarm2fs_ftruncate(const char *path, off_t size,
-                         struct fuse_file_info *fi)
+		   struct fuse_file_info *fi)
 {
 	gfarm_error_t e;
 
 	(void) path;
 	e = gfs_pio_truncate(get_filep(fi), size);
+	gfarm2fs_check_error(GFARM_MSG_2000024, OP_FTRUNCATE,
+				"gfs_pio_ftruncate", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -414,11 +589,13 @@ gfarm2fs_utime(const char *path, struct utimbuf *buf)
 
 	if (buf != NULL) {
 		gt[0].tv_sec = buf->actime;
-		gt[0].tv_nsec= 0;
+		gt[0].tv_nsec = 0;
 		gt[1].tv_sec = buf->modtime;
-		gt[1].tv_nsec= 0;
+		gt[1].tv_nsec = 0;
 	}
 	e = gfs_utimes(path, gt);
+	gfarm2fs_check_error(GFARM_MSG_2000025, OP_UTIME,
+				"gfs_utime", path, e);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -428,9 +605,15 @@ gfs_hook_open_flags_gfarmize(int open_flags)
 	int gfs_flags;
 
 	switch (open_flags & O_ACCMODE) {
-	case O_RDONLY:	gfs_flags = GFARM_FILE_RDONLY; break;
-	case O_WRONLY:	gfs_flags = GFARM_FILE_WRONLY; break;
-	case O_RDWR:	gfs_flags = GFARM_FILE_RDWR; break;
+	case O_RDONLY:
+		gfs_flags = GFARM_FILE_RDONLY;
+		break;
+	case O_WRONLY:
+		gfs_flags = GFARM_FILE_WRONLY;
+		break;
+	case O_RDWR:
+		gfs_flags = GFARM_FILE_RDWR;
+		break;
 	default: return (-1);
 	}
 
@@ -462,8 +645,11 @@ gfarm2fs_create(const char *path, mode_t mode, struct fuse_file_info *fi)
 
 	flags = gfs_hook_open_flags_gfarmize(fi->flags);
 	e = gfs_pio_create(path, flags, mode & GFARM_S_ALLPERM, &gf);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000026, OP_CREATE,
+					"gfs_pio_create", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	fi->fh = (unsigned long) gf;
 	return (0);
@@ -478,8 +664,11 @@ gfarm2fs_open(const char *path, struct fuse_file_info *fi)
 
 	flags = gfs_hook_open_flags_gfarmize(fi->flags);
 	e = gfs_pio_open(path, flags, &gf);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000027, OP_OPEN,
+					"gfs_pio_open", path, e);
 		return (-gfarm_error_to_errno(e));
+	}
 
 	fi->fh = (unsigned long) gf;
 	return (0);
@@ -495,8 +684,15 @@ gfarm2fs_read(const char *path, char *buf, size_t size, off_t offset,
 
 	(void) path;
 	e = gfs_pio_seek(get_filep(fi), offset, GFARM_SEEK_SET, &off);
-	if (e == GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000028, OP_READ,
+					"gfs_pio_seek", path, e);
+	} else {
 		e = gfs_pio_read(get_filep(fi), buf, size, &rv);
+		gfarm2fs_check_error(GFARM_MSG_2000029, OP_READ,
+					"gfs_pio_read", path, e);
+	}
+
 	if (e != GFARM_ERR_NO_ERROR)
 		rv = -gfarm_error_to_errno(e);
 
@@ -513,8 +709,15 @@ gfarm2fs_write(const char *path, const char *buf, size_t size,
 
 	(void) path;
 	e = gfs_pio_seek(get_filep(fi), offset, GFARM_SEEK_SET, &off);
-	if (e == GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000030, OP_WRITE,
+					"gfs_pio_seek", path, e);
+	} else {
 		e = gfs_pio_write(get_filep(fi), buf, size, &rv);
+		gfarm2fs_check_error(GFARM_MSG_2000031, OP_WRITE,
+					"gfs_pio_write", path, e);
+	}
+
 	if (e != GFARM_ERR_NO_ERROR)
 		rv = -gfarm_error_to_errno(e);
 
@@ -528,9 +731,11 @@ gfarm2fs_statfs(const char *path, struct statvfs *stbuf)
 	gfarm_off_t used, avail, files;
 
 	e = gfs_statfs(&used, &avail, &files);
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000032, OP_STATFS,
+					"gfs_statfs", path, e);
 		return (-gfarm_error_to_errno(e));
-	else {
+	} else {
 		stbuf->f_bsize = 1024;	/* XXX */
 		stbuf->f_frsize = 1024;	/* XXX */
 		stbuf->f_blocks = used + avail;
@@ -549,8 +754,13 @@ gfarm2fs_statfs(const char *path, struct statvfs *stbuf)
 static int
 gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 {
+	gfarm_error_t e;
+
 	(void) path;
-	gfs_pio_close(get_filep(fi));
+	e = gfs_pio_close(get_filep(fi));
+	gfarm2fs_check_error(GFARM_MSG_2000033, OP_RELEASE,
+				"gfs_pio_close", path, e);
+	gfarm2fs_replicate(path, fi);
 
 	return (0);
 }
@@ -561,10 +771,15 @@ gfarm2fs_fsync(const char *path, int isdatasync, struct fuse_file_info *fi)
 	gfarm_error_t e;
 
 	(void) path;
-	if (isdatasync)
+	if (isdatasync) {
 		e = gfs_pio_datasync(get_filep(fi));
-	else
+		gfarm2fs_check_error(GFARM_MSG_2000034, OP_FSYNC,
+					"gfs_pio_datasync", path, e);
+	} else {
 		e = gfs_pio_sync(get_filep(fi));
+		gfarm2fs_check_error(GFARM_MSG_2000035, OP_FSYNC,
+					"gfs_pio_sync", path, e);
+	}
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -574,8 +789,11 @@ gfarm2fs_setxattr(const char *path, const char *name, const char *value,
 	size_t size, int flags)
 {
 	gfarm_error_t e;
+
 	e = gfs_setxattr(path, name, value, size, flags);
-	return -gfarm_error_to_errno(e);
+	gfarm2fs_check_error(GFARM_MSG_2000036, OP_SETXATTR,
+				"gfs_setxattr", path, e);
+	return (-gfarm_error_to_errno(e));
 }
 
 static int
@@ -590,18 +808,22 @@ gfarm2fs_getxattr(const char *path, const char *name, char *value, size_t size)
 		 * NOTE: man getxattr(2) says that ENOATTR must be returned,
 		 * but it's not defined in header files.
 		 * We return -ENODATA because "strace ls -l /" is below.
-		 *		open("/", O_RDONLY|O_NONBLOCK|O_LARGEFILE|O_DIRECTORY) = 3
-		 *		....
-		 *		getxattr("/etc", "system.posix_acl_access"..., 0x0, 0) = -1 ENODATA (No data available)
-		 *		getxattr("/etc", "system.posix_acl_default"..., 0x0, 0) = -1 ENODATA (No data available)
-		 *  	...
+		 *   open("/", O_RDONLY|O_NONBLOCK|O_LARGEFILE|O_DIRECTORY) = 3
+		 *   ....
+		 *   getxattr("/etc", "system.posix_acl_access"..., 0x0, 0)
+		 *     = -1 ENODATA (No data available)
+		 *   getxattr("/etc", "system.posix_acl_default"..., 0x0, 0)
+		 *     = -1 ENODATA (No data available)
+		 *   ...
 		 */
-		return -ENODATA;
+		return (-ENODATA);
 	}
-	if (e != GFARM_ERR_NO_ERROR)
-		return -gfarm_error_to_errno(e);
-	else
-		return s;
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000037, OP_GETXATTR,
+				     "gfs_getxattr", path, e);
+		return (-gfarm_error_to_errno(e));
+	} else
+		return (s);
 }
 
 static int
@@ -611,21 +833,26 @@ gfarm2fs_listxattr(const char *path, char *list, size_t size)
 	size_t s = size;
 
 	e = gfs_listxattr(path, list, &s);
-	if (e == GFARM_ERR_NO_ERROR)
-		return s;
-	else
-		return -gfarm_error_to_errno(e);
+	if (e != GFARM_ERR_NO_ERROR) {
+		gfarm2fs_check_error(GFARM_MSG_2000038, OP_LISTXATTR,
+					"gfs_listxattr", path, e);
+		return (-gfarm_error_to_errno(e));
+	} else
+		return (s);
 }
 
 static int
 gfarm2fs_removexattr(const char *path, const char *name)
 {
 	gfarm_error_t e;
+
 	e = gfs_removexattr(path, name);
+	gfarm2fs_check_error(GFARM_MSG_2000039, OP_REMOVEXATTR,
+				"gfs_removexattr", path, e);
 	if (e == GFARM_ERR_NO_SUCH_OBJECT)
-		return -ENODATA;
+		return (-ENODATA);
 	else
-		return -gfarm_error_to_errno(e);
+		return (-gfarm_error_to_errno(e));
 }
 #endif /* HAVE_SYS_XATTR_H && ENABLE_XATTR */
 
@@ -664,7 +891,7 @@ static struct fuse_operations gfarm2fs_oper = {
     .setxattr	= gfarm2fs_setxattr,
     .getxattr	= gfarm2fs_getxattr,
     .listxattr	= gfarm2fs_listxattr,
-    .removexattr= gfarm2fs_removexattr,
+    .removexattr = gfarm2fs_removexattr,
 #endif
 };
 
@@ -677,8 +904,13 @@ uncache_parent(const char *path)
 {
 	char *p = strdup(path), *b;
 
-	if (p == NULL) /* XXX should report an error */
+	if (p == NULL) {
+		gflog_error(GFARM_MSG_2000040, \
+			    "strdup(%s): Cannot uncached parent dir " \
+			    "because not enough memory", \
+			    path);
 		return;
+	}
 
 	b = (char *)gfarm_path_dir_skip(p); /* UNCONST */
 	if (b > p && b[-1] == '/') {
@@ -798,7 +1030,7 @@ gfarm2fs_truncate_cached(const char *path, off_t size)
 
 static int
 gfarm2fs_ftruncate_cached(const char *path, off_t size,
-                         struct fuse_file_info *fi)
+			struct fuse_file_info *fi)
 {
 	int rv = gfarm2fs_ftruncate(path, size, fi);
 
@@ -918,7 +1150,7 @@ static struct fuse_operations gfarm2fs_cached_oper = {
     .setxattr	= gfarm2fs_setxattr_cached,
     .getxattr	= gfarm2fs_getxattr,
     .listxattr	= gfarm2fs_listxattr,
-    .removexattr= gfarm2fs_removexattr_cached,
+    .removexattr = gfarm2fs_removexattr_cached,
 #endif
 };
 
@@ -926,7 +1158,6 @@ static struct fuse_operations gfarm2fs_cached_oper = {
  *** main
  ***/
 
-
 #ifdef HAVE_GFARM_SCHEDULE_CACHE_DUMP
 void
 debug_handler(int signo)
@@ -942,10 +1173,6 @@ setup_dumper(void)
 #ifdef HAVE_GFARM_SCHEDULE_CACHE_DUMP
 	struct sigaction sa;
 
-	gflog_set_identifier(program_name);
-	gflog_auth_set_verbose(1);
-	gflog_syslog_open(LOG_PID, GFARM_DEFAULT_FACILITY);
-
 	sigemptyset(&sa.sa_mask);
 	sa.sa_handler = debug_handler;
 	sa.sa_flags = SA_RESTART;
@@ -953,13 +1180,128 @@ setup_dumper(void)
 #endif
 }
 
+/* KEY_GFARM2FS_OPT uses in "template has a format" case.
+ * This has no meaning except for just marker. */
+enum {
+	KEY_GFARM2FS_OPT,
+	KEY_F,
+	KEY_D,
+	KEY_VERSION,
+	KEY_HELP,
+};
 
-int main(int argc, char *argv[])
+#define GFARM2FS_OPT(t, p, v) \
+	{ t, offsetof(struct gfarm2fs_param, p), v }
+
+static struct fuse_opt gfarm2fs_opts[] = {
+	GFARM2FS_OPT("gfs_stat_timeout=%lf", cache_timeout, KEY_GFARM2FS_OPT),
+	GFARM2FS_OPT("-E %lf", cache_timeout, KEY_GFARM2FS_OPT),
+	/* GFARM2FS_OPT("use_stderr", use_syslog, 0), */
+	GFARM2FS_OPT("syslog=%s", facility, KEY_GFARM2FS_OPT),
+	GFARM2FS_OPT("loglevel=%s", loglevel, KEY_GFARM2FS_OPT),
+	GFARM2FS_OPT("ncopy=%d", ncopy, KEY_GFARM2FS_OPT),
+	GFARM2FS_OPT("copy_limit=%d", copy_limit, KEY_GFARM2FS_OPT),
+	FUSE_OPT_KEY("-f", KEY_F),
+	FUSE_OPT_KEY("-d", KEY_D),
+	FUSE_OPT_KEY("debug", KEY_D),
+	FUSE_OPT_KEY("-V", KEY_VERSION),
+	FUSE_OPT_KEY("--version", KEY_VERSION),
+	FUSE_OPT_KEY("-h", KEY_HELP),
+	FUSE_OPT_KEY("--help", KEY_HELP),
+	FUSE_OPT_END
+};
+
+static void
+usage(const char *progname)
+{
+	fprintf(stderr,
+"usage: %s mountpoint [options]\n"
+"\n"
+"general options:\n"
+"    -o opt,[opt...]         mount options\n"
+"    -h   --help             print help\n"
+"    -V   --version          print version\n"
+"\n"
+"GFARM2FS options:\n"
+"    -o syslog=facility      syslog facility (default: %s)\n"
+"    -o loglevel=priority    syslog priority level (default: %s)\n"
+"    -E T                    cache timeout for gfs_stat (default: 1.0 sec.)\n"
+"    -o gfs_stat_timeout=T   same -E option\n"
+"    -o ncopy=N              number of copies (default: 0)\n"
+"    -o copy_limit=N         maximum number of concurrent copy creations\n"
+"                            (default: 10)\n"
+		"\n", progname,
+		GFARM2FS_SYSLOG_FACILITY_DEFAULT,
+		GFARM2FS_SYSLOG_PRIORITY_DEFAULT);
+}
+
+static int
+gfarm2fs_fuse_main(struct fuse_args *args, struct fuse_operations *fo)
+{
+#if FUSE_VERSION >= 26
+	return (fuse_main(args->argc, args->argv, fo, NULL));
+#else
+	return (fuse_main(args->argc, args->argv, fo));
+#endif
+}
+
+static int
+gfarm2fs_opt_proc(void *data, const char *arg, int key,
+			struct fuse_args *outargs)
+{
+	struct gfarm2fs_param *paramsp = data;
+
+	switch (key) {
+	case FUSE_OPT_KEY_OPT: /* -?, -o opt, --opt */
+		return (1); /* through */
+	case FUSE_OPT_KEY_NONOPT:
+		if (!paramsp->mount_point)
+			paramsp->mount_point = arg;
+		return (1); /* through */
+	case KEY_F:
+		paramsp->foreground = 1;
+		return (1); /* through */
+	case KEY_D:
+		paramsp->debug = 1;
+		return (1); /* through */
+	case KEY_VERSION:
+		fprintf(stderr, "GFARM2FS version %s\n", VERSION);
+#if FUSE_VERSION >= 25
+		fuse_opt_add_arg(outargs, "--version");
+		gfarm2fs_fuse_main(outargs, NULL);
+#endif
+		exit(0);
+	case KEY_HELP:
+		usage(outargs->argv[0]);
+		fuse_opt_add_arg(outargs, "-ho");
+		gfarm2fs_fuse_main(outargs, NULL);
+		exit(1);
+	default:
+		return (0);
+	}
+}
+
+int
+main(int argc, char *argv[])
 {
 	struct fuse_operations *operation_mode = &gfarm2fs_cached_oper;
 	gfarm_error_t e;
-	struct fuse_args args = FUSE_ARGS_INIT(0, NULL);
-	int i;
+	int ret_fuse_main;
+	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
+	int syslog_priority;
+	int syslog_facility = -1;
+
+	struct gfarm2fs_param params = {
+		.mount_point = NULL,
+		.foreground = 0,
+		.debug = 0,
+		.cache_timeout = -1.0,
+		.use_syslog = 1,
+		.facility = NULL,
+		.loglevel = NULL,
+		.ncopy = 0,
+		.copy_limit = 10
+	};
 
 	umask(0);
 	e = gfarm_initialize(&argc, &argv);
@@ -968,51 +1310,77 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-	if (argc > 0)
-		fuse_opt_add_arg(&args, argv[0]);
-
+	if (fuse_opt_parse(&args, &params, gfarm2fs_opts,
+			   gfarm2fs_opt_proc) == -1) {
+		fprintf(stderr, "failed to parse option\n");
+		exit(1);
+	}
 	/*
 	 * specify '-s' option to disable multithreaded operations
 	 * libgfarm is not thread-safe for now
 	 */
 	fuse_opt_add_arg(&args, "-s");
 
-	for (i = 1; i < argc; i++) {
-		if (argv[i][0] == '-' && argv[i][1] == 'E') {
-			double timeout;
-			char *ep;
+	if (params.mount_point == NULL) {
+		fprintf(stderr, "missing mountpoint\n");
+		fprintf(stderr, "see `%s -h' for usage\n", program_name);
+		exit(1);
+	}
+	mount_point = params.mount_point;
 
-			if (i + 1 >= argc) {
-				fprintf(stderr, "%s: -E: missing <seconds>\n",
-				    program_name);
-				exit(EXIT_FAILURE);
-			}
-			errno = 0;
-			timeout = strtod(argv[i + 1], &ep);
-			if (ep == optarg || *ep != '\0') {
-				fprintf(stderr, "%s: -E %s: %s\n",
-				    program_name, argv[i+1],
-				    "invalid argument");
-				exit(EXIT_FAILURE);
-			} else if (errno != 0) {
-				fprintf(stderr, "%s: -E %s: %s\n",
-				    program_name, argv[i+1], strerror(errno));
-				exit(EXIT_FAILURE);
-			}
-			i++;
+	if (params.foreground || params.debug) {
+		params.use_syslog = 0; /* use stderr */
+		if (params.loglevel == NULL)
+			params.loglevel = GFARM2FS_SYSLOG_PRIORITY_DEBUG;
+	}
 
-			if (timeout == 0.0) {
-				gfs_stat_cache_enable(0); /* disable cache */
-				operation_mode = &gfarm2fs_oper;
-			} else {
-				gfs_stat_cache_expiration_set(timeout*1000.0);
-			}
-		} else {
-			fuse_opt_add_arg(&args, argv[i]);
+	if (params.loglevel == NULL)
+		params.loglevel = GFARM2FS_SYSLOG_PRIORITY_DEFAULT;
+	syslog_priority = gflog_syslog_name_to_priority(params.loglevel);
+	if (syslog_priority == -1) {
+		fprintf(stderr, "invalid loglevel: `%s'\n", params.loglevel);
+		fprintf(stderr, "see `%s -h' for usage\n", program_name);
+		exit(1);
+	}
+	gflog_set_priority_level(syslog_priority);
+
+	gflog_set_identifier(program_name);
+	gflog_auth_set_verbose(1);
+
+	if (params.use_syslog) {
+		if (params.facility == NULL)
+			params.facility = GFARM2FS_SYSLOG_FACILITY_DEFAULT;
+		syslog_facility = gflog_syslog_name_to_facility(
+			params.facility);
+		if (syslog_facility == -1) {
+			fprintf(stderr, "invalid facility: `%s'\n",
+				params.facility);
+			fprintf(stderr, "see `%s -h' for usage\n",
+				program_name);
+			exit(1);
 		}
 	}
 
+	if (params.cache_timeout > 0.0) {
+		gfs_stat_cache_expiration_set(params.cache_timeout*1000.0);
+	} else if (params.cache_timeout == 0.0) {
+		gfs_stat_cache_enable(0); /* disable cache */
+		operation_mode = &gfarm2fs_oper;
+	}
+
+	/* end of setting params */
+
+	gfarm2fs_replicate_init(&params);
+
 	setup_dumper();
 
-	return (fuse_main(args.argc, args.argv, operation_mode));
+	if (params.use_syslog) /* just before fuse_main */
+		gflog_syslog_open(LOG_PID, syslog_facility);
+
+	ret_fuse_main = gfarm2fs_fuse_main(&args, operation_mode);
+	fuse_opt_free_args(&args);
+
+	gfarm2fs_replicate_final();
+
+	return (ret_fuse_main);
 }
